{-+ PFE command line parsing utilities
-}
module PfeParse where
import Char(isUpper)
import Monad(when)

import HsName(ModuleName(..),sameModuleName,parseModuleName,isMainModule)
import TypedIds(NameSpace(..))

import PFE0(getCurrentModuleGraph)

import PrettyPrint(pp,(<+>),fsep)
import CmdLineParser(Cmd(..))
import MUtils((@@),( # ),( <# ),concatMapM,swap,apBoth)

moduleArg f = moduleArgs (mapM_ f)
moduleArgs f = Args "<modules>" (f @@ concatMapM checkModuleName)

checkModuleName s =
  do ms <- filter sameModule . map (fst.snd) # getCurrentModuleGraph
     when (null ms) $ fail (show s++" "++show m++show ms++": unknown module")
     return ms
  where
    m = parseModuleName s
    sameModule = if isMainModule m then (==) m else sameModuleName s
    -- "Main{-file.hs-}" selects one particular Main module,
    -- "Main" select all main modules in a project

just ms = if null ms then Nothing else Just ms

qualIds f = Args "<M1.x1 ... Mn.xn>" (f @@ parseQualIds)

qualId f = Args "<M.x>" (f @@ parseOneQualId)

parseQualIds = mapM parseQualId
parseOneQualId = parseQualId @@ one
  where
     one [q] = return q
     one qs = fail $ "Exactly one qualified name is required: "++unwords qs

parseQualId s =
    case splitQualName s of
      Just (m,n) -> (,) # checkModuleName1 m <# return n
             -- TODO: also check that m.n is defined!
      _ -> fail $ "Qaulified name required: "++s
  where
    splitQualName = fmap (apBoth reverse . swap) . split . reverse

    split s = case break (=='.') s of
		     (s1,'.':s2) -> Just (s1,s2)
		     _ -> Nothing
{-
    isQual s =
      case break (=='.') s of
	(c:_,'.':_:_) -> isUpper c
	_ -> False
-}
    checkModuleName1 = one @@ checkModuleName
      where
     one [q] = return q
     one qs = fail $ pp $ "Ambiguous module name:"<+>fsep qs

entityId f = Args "[type|class|value] <M.x>" (f @@ parseEntId)
  where
    parseEntId args0 = (,) ns # parseOneQualId args1
       where
         (ns,args1) =
           case args0 of
	     arg:args | isClassOrType arg -> (Just ClassOrTypeNames,args)
		      | isValue arg       -> (Just ValueNames,args)
	     _ -> (Nothing,args0)
         isClassOrType arg = arg `elem` ["type","class"]
	 isValue arg = arg `elem` ["value","con"]
